use bevy::prelude::*;
use bevy_egui::{egui, EguiContexts, EguiPlugin};
use my_library::RandomNumberGenerator; //(1) //(2)

#[derive(Clone, Copy, PartialEq, Eq, Debug, Hash, States, Default)] //(3)
enum GamePhase {
  #[default]
  Player,
  Cpu,
}

fn main() {
  App::new()
  .add_plugins(DefaultPlugins)
  .add_plugins(EguiPlugin)
  .add_systems(Startup, setup) //(4)
  .add_state::<GamePhase>() //(5)
  .add_systems(Update, display_score) //(6)
  .add_systems(Update, player.run_if(
    in_state(GamePhase::Player))) //(7)
  .add_systems(Update, cpu.run_if(
    in_state(GamePhase::Cpu))) //(8)
  .run();
}

#[derive(Resource)]
struct GameAssets {
  dice: Handle<TextureAtlas>, //(9)
}

#[derive(Clone, Copy, Resource)]
struct Scores {
  player: usize,//(10)
  cpu: usize,
}

#[derive(Component)] //(11)
struct HandDie;

#[derive(Resource)]
struct Random(RandomNumberGenerator);//(12)

#[derive(Resource)]
struct HandTimer(Timer);//(13)

fn setup(
  asset_server: Res<AssetServer>, //(14)
  mut texture_atlases: ResMut<Assets<TextureAtlas>>, //(15)
  mut commands: Commands, //(16)
) {
  commands.spawn(Camera2dBundle::default()); //(17)
  let atlas = TextureAtlas::from_grid(
    //(18)
    asset_server.load("dice.png"), //(19)
    Vec2::new(52.0, 52.0), //(20)
    6,                     //(21)
    1,                     //(22)
    None,
    None
  );
  let atlas_handle = texture_atlases.add(atlas); //(23)
  commands.insert_resource(GameAssets { dice: atlas_handle }); //(24)
  commands.insert_resource(Scores { cpu: 0, player: 0 }); //(25)
  commands.insert_resource(Random(RandomNumberGenerator::new())); //(26)
  commands.insert_resource(HandTimer(Timer::from_seconds(0.5, TimerMode::Repeating))); //(27)
}

fn display_score(
  scores: Res<Scores>,
  mut egui_context: EguiContexts, //(28)
) {
  egui::Window::new("Total Scores").show(egui_context.ctx_mut(), |ui| {
    //(29)
    ui.label(&format!("Player: {}", scores.player)); //(30)
    ui.label(&format!("CPU: {}", scores.cpu));
  });
}

fn clear_die(
  hand_query: &Query<(Entity, &TextureAtlasSprite), With<HandDie>>,
  commands: &mut Commands,
) {
  hand_query
    .iter()
    .for_each(|(entity, _)| commands.entity(entity).despawn());
}

fn spawn_die(
  hand_query: &Query<(Entity, &TextureAtlasSprite), With<HandDie>>, //(31)
  commands: &mut Commands,
  assets: &GameAssets,
  new_roll: usize,
  color: Color,
) {
  let rolled_die = hand_query.iter().count() as f32 * 52.0; //(32)
  commands
    .spawn(SpriteSheetBundle {
      //(33)
      texture_atlas: assets.dice.clone(),
      transform: Transform::from_xyz(rolled_die - 400.0, 60.0, 1.0), //(34)
      sprite: TextureAtlasSprite {
        index: new_roll - 1, //(35)
        color,
        ..default()
      },
      ..default()
    })
    .insert(HandDie);
}

fn player(
  hand_query: Query<(Entity, &TextureAtlasSprite), With<HandDie>>,
  mut commands: Commands,
  mut rng: ResMut<Random>,
  assets: Res<GameAssets>,
  mut scores: ResMut<Scores>,
  mut state: ResMut<NextState<GamePhase>>,
  mut egui_context: EguiContexts,
) {
  egui::Window::new("Play Options").show(egui_context.ctx_mut(), |ui| {
    let hand_score: usize =
      hand_query.iter().map(|(_, ts)| ts.index + 1).sum();//(36)
    ui.label(&format!("Score for this hand: {hand_score}"));

    if ui.button("Roll Dice").clicked() {//(37)
      let new_roll = rng.0.range(1..=6);
      if new_roll == 1 {
        // End turn!
        clear_die(&hand_query, &mut commands);
        state.set(GamePhase::Cpu);//(38)
      } else {
        spawn_die(
          &hand_query,
          &mut commands,
          &assets,
          new_roll,
          Color::WHITE,
        );
      }
    }
    if ui.button("Pass - Keep Hand Score").clicked() {
      let hand_total: usize =
        hand_query.iter().map(|(_, ts)| ts.index + 1).sum();
      scores.player += hand_total;
      clear_die(&hand_query, &mut commands);
      state.set(GamePhase::Cpu);
    }
  });
}

#[allow(clippy::too_many_arguments)]
fn cpu(
  hand_query: Query<(Entity, &TextureAtlasSprite), With<HandDie>>,
  mut state: ResMut<NextState<GamePhase>>,
  mut scores: ResMut<Scores>,
  mut rng: ResMut<Random>,
  mut commands: Commands,
  assets: Res<GameAssets>,
  mut timer: ResMut<HandTimer>,
  time: Res<Time>,
) {
  timer.0.tick(time.delta());//(39)
  if timer.0.just_finished() {//(40)
    let hand_total: usize =
      hand_query.iter().map(|(_, ts)| ts.index + 1).sum();
    if hand_total < 20 && scores.cpu + hand_total < 100 {
      let new_roll = rng.0.range(1..=6);
      if new_roll == 1 {
        clear_die(&hand_query, &mut commands);
        state.set(GamePhase::Player);
      } else {
        spawn_die(
          &hand_query,
          &mut commands,
          &assets,
          new_roll,
          Color::BLUE,
        );
      }
    } else {
      scores.cpu += hand_total;//(41)
      state.set(GamePhase::Player);
      hand_query
        .iter()
        .for_each(|(entity, _)| commands.entity(entity).despawn());
    }
  }
}
