use rand::{prelude::StdRng, Rng, SeedableRng};
use std::ops::Range;

pub struct RandomNumberGenerator {
  rng: StdRng,
}

impl Default for RandomNumberGenerator {
  fn default() -> Self {
    Self::new()
  }
}

impl RandomNumberGenerator {
  pub fn new() -> Self {
    Self {
      rng: StdRng::from_entropy(),
    }
  }

  pub fn seeded(seed: u64) -> Self {
    Self {
      rng: StdRng::seed_from_u64(seed),
    }
  }

  pub fn next<T>(&mut self) -> T
  where rand::distributions::Standard: rand::prelude::Distribution<T>
  {
    self.rng.gen()
  }

  pub fn range<T>(&mut self, range: Range<T>) -> T
  where
  T: rand::distributions::uniform::SampleUniform + PartialOrd,
  {
    self.rng.gen_range(range)
  }
}

#[cfg(test)]
mod test {
  use super::*;

  #[test]
  fn test_range_bounds() {
    let mut rng = RandomNumberGenerator::new();
    for _ in 0..1000 {
      let n = rng.range(1..10);
      assert!(n >= 1);
      assert!(n < 10);
    }
  }

  #[test]
  fn test_reproducibility() {
    let mut rng = (
      RandomNumberGenerator::seeded(1),
      RandomNumberGenerator::seeded(1),
    );
    (0..1000).for_each(|_| {
      assert_eq!(
        rng.0.range(u32::MIN..u32::MAX),
        rng.1.range(u32::MIN..u32::MAX),
      );
    });
  }

  #[test]
  fn test_next_types() {
    let mut rng = RandomNumberGenerator::new();
    let _ : i32 = rng.next();//(1)
    let _ = rng.next::<f32>();//(2)
  }

  #[test]
  fn test_float() {
    let mut rng = RandomNumberGenerator::new();
    for _ in 0..1000 {
      let n = rng.range(-5000.0f32..5000.0f32);
      assert!(n.is_finite());
      assert!(n > -5000.0);
      assert!(n < 5000.0);
    }
  }
}
