use crate::AssetStore;
use bevy::prelude::*;

#[derive(Clone)]
pub enum AssetType {
  Image,
  Sound,
}

#[derive(Resource, Clone)]// (1)
pub struct AssetManager {
  asset_list: Vec<(String, String, AssetType)>, // (2)
}

impl AssetManager {
  pub fn new() -> Self {
    Self {
      asset_list: vec![
        ("main_menu".to_string(), "main_menu.png".to_string(), 
          AssetType::Image),
        ("game_over".to_string(), "game_over.png".to_string(), 
          AssetType::Image),
      ],
    }
  }

  fn asset_exists(filename: &str) -> anyhow::Result<()> {
    #[cfg(not(target_arch = "wasm32"))]
    {
      let current_directory = std::env::current_dir()?;
      let assets = current_directory.join("assets");
      let new_image = assets.join(filename);
      if !new_image.exists() {
        return Err(anyhow::Error::msg(format!(
          "{} not found in assets directory",
          &filename
        )));
      }
    }
    Ok(())
  }

  pub fn add_image<S: ToString>(
    mut self,
    tag: S,
    filename: S,
  ) -> anyhow::Result<Self> {
    let filename = filename.to_string();
    AssetManager::asset_exists(&filename)?;
    self
      .asset_list
      .push((tag.to_string(), filename, AssetType::Image));
    Ok(self)
  }

  pub fn add_sound<S: ToString>(
    mut self,
    tag: S,
    filename: S,
  ) -> anyhow::Result<Self> {
    let filename = filename.to_string();
    AssetManager::asset_exists(&filename)?;
    self
      .asset_list
      .push((tag.to_string(), filename, AssetType::Sound));
    Ok(self)
  }
}

impl Plugin for AssetManager {
  fn build(&self, app: &mut bevy::prelude::App) {
    app.insert_resource(self.clone());
    app.add_systems(Startup, setup);
  }
}

fn setup(
  asset_resource: Res<AssetManager>,//(3)
  mut commands: Commands,
  asset_server: Res<AssetServer>,
) {
  let mut assets = AssetStore {//(4)
    asset_index: bevy::utils::HashMap::new(),
  };
  asset_resource.asset_list.iter().for_each(
    |(tag, filename, asset_type)| {
      match asset_type {
        _ => {
          // Most asset types don't require a separate loader
          assets
            .asset_index
            .insert(tag.clone(), asset_server.load_untyped(filename));
        }
      }
    },
  );
  commands.remove_resource::<AssetManager>();//(5)
  commands.insert_resource(assets);//(6)
}
