use bevy::{prelude::*, utils::HashMap, log};

pub enum AnimationOption {
    None,//(1)
    NextFrame, //(2)
    GoToFrame(usize),//(3)
    SwitchToAnimation(String),//(4)
    PlaySound(String),//(5)
}

pub struct AnimationFrame {
    sprite_index: usize,//(6)
    delay_ms: u128,//(7)
    action: Vec<AnimationOption>,//(8)
}

impl AnimationFrame {
    pub fn new(//(9)
               sprite_index: usize,
               delay_ms: u128,
               action: Vec<AnimationOption>
    ) -> Self {
        Self { sprite_index, delay_ms, action }
    }
}

pub struct PerFrameAnimation {
    pub frames: Vec<AnimationFrame>,
}

impl PerFrameAnimation {
    pub fn new(frames: Vec<AnimationFrame>) -> Self {
        Self { frames }
    }
}

#[derive(Resource)]
pub struct Animations(HashMap<String, PerFrameAnimation>);

impl Animations {
    pub fn new() -> Self {
        Self(HashMap::new())
    }

    pub fn with_animation<S: ToString>(
        mut self,
        tag: S,
        animation: PerFrameAnimation
    ) -> Self
    {
        self.0.insert(
            tag.to_string(),
            animation
        );
        self
    }
}

#[derive(Component)]
pub struct AnimationCycle {
    animation_tag: String,// (10)
    current_frame: usize,// (11)
    timer: u128,// (12)
}

impl AnimationCycle {
    pub fn new<S: ToString>(tag: S) -> Self {// (13)
        Self {
            animation_tag: tag.to_string(),
            current_frame: 0,
            timer: 0,
        }
    }

    pub fn switch<S: ToString>(&mut self, new: S) {// (14)
        let new = new.to_string();
        if new != self.animation_tag {// (15)
            self.animation_tag = new;
            self.current_frame = 0;
            self.timer = 0;
        }
    }
}

pub fn cycle_animations(
    animations: Res<Animations>,// (16)
    mut animated: Query<(&mut AnimationCycle, &mut TextureAtlasSprite)>,// (17)
    time: Res<Time>,// (18)
    assets: Res<crate::AssetStore>,// (19)
    mut commands: Commands,
    loaded_assets: Res<crate::LoadedAssets>,
) {
    let ms_since_last_call = time.delta().as_millis();// (20)
    animated.for_each_mut(|(mut animation, mut sprite)| {// (21)
        animation.timer += ms_since_last_call;// (22)
        if let Some(cycle) = animations.0.get(&animation.animation_tag) {// (23)
            let current_frame = &cycle.frames[animation.current_frame];// (24)
            if animation.timer > current_frame.delay_ms {// (25)
                animation.timer = 0;// (26)
                for action in current_frame.action.iter() {// (27)
                    match action {
                        AnimationOption::None => {},
                        AnimationOption::NextFrame => {// (28)
                            animation.current_frame += 1;
                        }
                        AnimationOption::GoToFrame(frame) => {// (29)
                            animation.current_frame = *frame;
                        }
                        AnimationOption::SwitchToAnimation(new) => {// (30)
                            animation.animation_tag = new.to_string();
                            animation.current_frame = 0;
                        }
                        AnimationOption::PlaySound(tag) => {// (31)
                            assets.play(tag, &mut commands, &loaded_assets);
                        }
                    }
                    sprite.index = cycle
                        .frames[animation.current_frame]
                        .sprite_index;// (32)
                }
            }
        } else {
            log::warn!("Animation Cycle [{}] not found!", animation.animation_tag);
        }
    });
}

#[macro_export]
macro_rules! spawn_animated_sprite {
   ($assets:expr, $commands:expr, $index:expr, $x:expr, $y:expr, $z:expr,
      $animation_name:expr, $($component:expr),*) =>
  {
      $commands.spawn(SpriteSheetBundle{
         texture_atlas: $assets.get_atlas_handle($index).unwrap(),
         transform: Transform::from_xyz($x, $y, $z),
         ..default()
      })
        .insert(AnimationCycle::new($animation_name))
      $(
        .insert($component)
      )*
   };
  }

#[derive(Component)]
pub struct ContinualParallax {
    image_width: f32, // (33)
    move_every_ms: u128, // (34)
    scroll_speed: Vec2, // (35)
    timer: u128, // (36)
}

impl ContinualParallax {
    pub fn new(image_width: f32, move_every_ms: u128, scroll_speed: Vec2)
               -> Self {
        Self {
            image_width, move_every_ms, scroll_speed, timer: 0
        }
    }
}

pub fn continual_parallax(
    mut animated: Query<(&mut ContinualParallax, &mut Transform)>,
    time: Res<Time>,
) {
    let ms_since_last_call = time.delta().as_millis();
    animated.for_each_mut(|(mut parallax, mut transform)| {
        parallax.timer += ms_since_last_call;
        if parallax.timer >= parallax.move_every_ms {
            parallax.timer = 0;
            transform.translation.x -= parallax.scroll_speed.x;
            transform.translation.y -= parallax.scroll_speed.y;
            if transform.translation.x <= (0.0 - parallax.image_width) {
                transform.translation.x = parallax.image_width;
            }
        }
    });
}
